%{ open Ast %}
/*Token keys as defined in the AST*/
%token SEMI LPAR RPAR LBRACE RBRACE COMMA NEWLINE
%token ADD SUB MULT DIV ADDASSN MINUSASSN MULTASSN DIVASSN ASSIGN SQUARE POWER FACTORIAL MODULUS INCREMENT DECREMENT
%token EQ NEQ LT LEQ GT GEQ
%token RETURN IF ELSE FOR INT ENDIF ENDFOR ENDWHILE WHILE
%token FN ENDFN
%token <int> LITERAL
%token <string> ID
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%right ADDASSN
%right MINUSASSN
%right DIVASSN
%right MULTASSN
%left EQ NEQ
%left LT GT LEQ GEQ
%left ADD SUB 
%left MULT DIV MODULUS
%left SQUARE
%left POWER FACTORIAL

%start program
%type <Ast.program> program

%%

program:
   /* nothing */ { [], [] }
 | program vdecl { ($2 :: fst $1), snd $1 }
 | program fdecl { fst $1, ($2 :: snd $1) }

fdecl:
   NEWLINE FN ID LPAR formals_opt RPAR NEWLINE vdecl_list stmt_list ENDFN NEWLINE
     { { fname = $3;
	 formals = $5;
	 locals = List.rev $8;
	 body = List.rev $9 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    ID                   { [$1] }
  | formal_list COMMA ID { $3 :: $1 }

vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
   INT ID NEWLINE { $2 }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
  expr NEWLINE { Expr($1) }
  | RETURN expr NEWLINE { Return($2) }
  | IF LPAR expr RPAR NEWLINE stmt_list %prec NOELSE ENDIF NEWLINE { If($3, Block(List.rev $6), Block([])) }
  | IF LPAR expr RPAR NEWLINE stmt_list ELSE NEWLINE stmt_list ENDIF NEWLINE   { If($3, Block(List.rev $6), Block(List.rev $9)) }
  | FOR LPAR expr SEMI expr SEMI expr RPAR  NEWLINE stmt_list ENDFOR NEWLINE
     { For($3, $5, $7, Block(List.rev $10)) }
  | WHILE LPAR expr RPAR NEWLINE stmt_list ENDWHILE NEWLINE { While($3, Block(List.rev $6)) }

/*MATCHING RULES FOR EXPRESSIONS*/
expr:
    LITERAL          { Literal($1) }
  | ID               { Id($1) }
  | expr ADD  expr { Binop($1, Add,   $3) }
  | expr SUB  expr { Binop($1, Sub,   $3) }
  | expr MULT  expr { Binop($1, Mult,  $3) }
  | expr DIV expr { Binop($1, Div,   $3) }
  | expr EQ     expr { Binop($1, Equal, $3) }
  | expr NEQ    expr { Binop($1, Neq,   $3) }
  | expr LT     expr { Binop($1, Less,  $3) }
  | expr LEQ    expr { Binop($1, Leq,   $3) }
  | expr GT     expr { Binop($1, Greater,  $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3) }
  | expr MODULUS expr {Binop($1, Mod, $3)}
  | expr SQUARE        { Binop($1, Mult, $1) }
  | expr FACTORIAL       { Binop($1, Fact, Literal(1)) }
  | expr POWER   expr { Binop($1, Pow, $3) }
  | ID INCREMENT       { Assign($1, Binop(Id($1), Add, Literal(1))) }
  | ID DECREMENT      { Assign($1, Binop(Id($1), Sub, Literal(1))) }
  | ID ADDASSN expr  { Assign($1, Binop(Id($1), Add, $3))}
  | ID MINUSASSN expr  { Assign($1, Binop(Id($1), Sub, $3))}
  | ID MULTASSN expr  { Assign($1, Binop(Id($1), Mult, $3))}
  | ID DIVASSN expr  { Assign($1, Binop(Id($1), Div, $3))}
  | ID ASSIGN expr   { Assign($1, $3) }
  | ID LPAR actuals_opt RPAR { Call($1, $3) }
  | LPAR expr RPAR { $2 }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }
